@@ -6,7 +6,7 @@ import logging
 
 from .samples import samples
 from .fwrapper import _FunctionWrapper
-from .autocorr import _autocorr_time
+from .autocorr import AutoCorrTime
 
 
 class sampler:
@@ -106,7 +106,6 @@ class sampler:
             raise ValueError("The total probability of all proposals must be equal to 1.0.")
 
 
-
     def run(self,
             start,
             nsteps=1000,
@@ -146,8 +145,8 @@ class sampler:
         self.thin = int(thin)
         self.samples.extend(self.nsteps//self.thin)
 
-        # Devine Number of Log Prob Evaluations vector
-        self.neval = np.zeros(self.nsteps)
+        # Devide Number of Log Prob Evaluations vector
+        self.neval = np.zeros(self.nsteps, dtype=int)
 
         # Define tuning count
         ncount = 0
@@ -214,18 +213,43 @@ class sampler:
                 # Initialise number of Log prob calls
                 ncall = 0
 
-                # Left stepping-out
+                # Left stepping-out initialisation
                 mask_J = np.full(int(self.nwalkers/2),True)
                 Z_L = np.empty(int(self.nwalkers/2))
                 X_L = np.empty((int(self.nwalkers/2),self.ndim))
 
+                # Right stepping-out initialisation
+                mask_K = np.full(int(self.nwalkers/2),True)
+                Z_R = np.empty(int(self.nwalkers/2))
+                X_R = np.empty((int(self.nwalkers/2),self.ndim))
+
                 cnt = 0
-                while len(mask_J[mask_J])>0:
+                # Stepping-Out procedure
+                while len(mask_J[mask_J])>0 or len(mask_K[mask_K])>0:
+                    if len(mask_J[mask_J])>0:
+                        cnt += 1
+                    if len(mask_K[mask_K])>0:
+                        cnt += 1
+                    if cnt > self.maxiter:
+                        raise RuntimeError('Number of expansions exceeded maximum limit! \n' +
+                                           'Make sure that the pdf is well-defined. \n' +
+                                           'Otherwise increase the maximum limit (maxiter=10^4 by default).')
+
                     for j in indeces[mask_J]:
                         if J[j] < 1:
                             mask_J[j] = False
+
+                    for j in indeces[mask_K]:
+                        if K[j] < 1:
+                            mask_K[j] = False
+
                     X_L[mask_J] = directions[mask_J] * L[mask_J][:,np.newaxis] + X[active][mask_J]
-                    Z_L[mask_J] = np.asarray(list(distribute(self.logprob_fn,X_L[mask_J])))
+                    X_R[mask_K] = directions[mask_K] * R[mask_K][:,np.newaxis] + X[active][mask_K]
+
+                    Z_LR_masked = np.asarray(list(distribute(self.logprob_fn, np.concatenate([X_L[mask_J],X_R[mask_K]]))))
+                    Z_L[mask_J] = Z_LR_masked[:X_L[mask_J].shape[0]]
+                    Z_R[mask_K] = Z_LR_masked[X_L[mask_J].shape[0]:]
+
                     for j in indeces[mask_J]:
                         ncall += 1
                         if Z0[j] < Z_L[j]:
@@ -234,24 +258,7 @@ class sampler:
                             nexp += 1
                         else:
                             mask_J[j] = False
-                    cnt += 1
-                    if cnt > self.maxiter:
-                        raise RuntimeError('Number of expansions exceeded maximum limit! \n' +
-                                           'Make sure that the pdf is well-defined. \n' +
-                                           'Otherwise increase the maximum limit (maxiter=10^4 by default).')
 
-                # Right stepping-out
-                mask_K = np.full(int(self.nwalkers/2),True)
-                Z_R = np.empty(int(self.nwalkers/2))
-                X_R = np.empty((int(self.nwalkers/2),self.ndim))
-
-                cnt = 0
-                while len(mask_K[mask_K])>0:
-                    for j in indeces[mask_K]:
-                        if K[j] < 1:
-                            mask_K[j] = False
-                    X_R[mask_K] = directions[mask_K] * R[mask_K][:,np.newaxis] + X[active][mask_K]
-                    Z_R[mask_K] = np.asarray(list(distribute(self.logprob_fn,X_R[mask_K])))
                     for j in indeces[mask_K]:
                         ncall += 1
                         if Z0[j] < Z_R[j]:
@@ -260,11 +267,7 @@ class sampler:
                             nexp += 1
                         else:
                             mask_K[j] = False
-                    cnt += 1
-                    if cnt > self.maxiter:
-                        raise RuntimeError('Number of expansions exceeded maximum limit! \n' +
-                                           'Make sure that the pdf is well-defined. \n' +
-                                           'Otherwise increase the maximum limit (maxiter=10^4 by default).')
+
 
                 # Shrinking procedure
                 Widths = np.empty(int(self.nwalkers/2))
@@ -330,6 +333,13 @@ class sampler:
         # Close progress bar
         if progress:
             t.close()
+    
+
+    def run_mcmc(self, *args, **kwargs):
+        """
+        Wrapper for run method.
+        """
+        return self.run(*args, **kwargs)
 
 
     def reset(self):
@@ -337,6 +347,42 @@ class sampler:
         Reset the state of the sampler. Delete any samples stored in memory.
         """
         self.samples = samples(self.ndim, self.nwalkers)
+    
+
+    def get_chain(self, flat=False, thin=1, discard=0):
+        """
+        Get the Markov chain containing the samples.
+
+        Args:
+            flat (bool) : If True then flatten the chain into a 2D array by combining all walkers (default is False).
+            thin (int) : Thinning parameter (the default value is 1).
+            discard (int) : Number of burn-in steps to be removed from each walker (default is 0).
+
+        Returns:
+            Array object containg the Markov chain samples (2D if flat=True, 3D if flat=False).
+        """
+        if flat:
+            return self.samples.flatten(discard=discard, thin=thin)
+        else:
+            return self.chain[discard::thin,:,:]
+
+    
+    def get_log_prob(self, flat=False, thin=1, discard=0):
+        """
+        Get the value of the log probability function evalutated at the samples of the Markov chain.
+
+        Args:
+            flat (bool) : If True then flatten the chain into a 1D array by combining all walkers (default is False).
+            thin (int) : Thinning parameter (the default value is 1).
+            discard (int) : Number of burn-in steps to be removed from each walker (default is 0).
+
+        Returns:
+            Array containing the value of the log probability at the samples of the Markov chain (1D if flat=True, 2D otherwise).
+        """
+        if flat:
+            return self.samples.flatten_logprob(discard=discard, thin=thin)
+        else:
+            return self.samples.logprob[discard::thin,:]
 
 
     @property
@@ -350,29 +396,15 @@ class sampler:
         return self.samples.chain
 
 
-    def flatten(self, discard=0, thin=1):
-        """
-        Flatten the chain.
-
-        Args:
-            discard (int): The number of burn-in steps to remove from each walker (default is 0).
-            thin (int): The ammount to thin the chain (default is 1, no thinning).
-
-        Returns:
-            2D Flattened chain.
-        """
-        return self.samples.flatten(discard, thin)
-
-
     @property
-    def autocorr_time(self):
+    def act(self):
         """
         Integrated Autocorrelation Time (IAT) of the Markov Chain.
 
         Returns:
             Array with the IAT of each parameter.
         """
-        return _autocorr_time(np.swapaxes(self.chain[int(self.nsteps/(self.thin*2.0)):,:,:], 0, 1))
+        return AutoCorrTime(self.chain[int(self.nsteps/(self.thin*2.0)):,:,:])
 
 
     @property
@@ -383,7 +415,7 @@ class sampler:
         Returns:
             ESS
         """
-        return self.nwalkers * self.samples.length / np.mean(self.autocorr_time)
+        return self.nwalkers * self.samples.length / np.mean(self.act)
 
 
     @property
@@ -418,6 +450,14 @@ class sampler:
         """
         return np.asarray(self.mus)
 
+    
+    @property
+    def get_last_sample(self):
+        """
+            Return the last position of the walkers.
+        """
+        return self.chain[-1]
+
 
     @property
     def summary(self):
@@ -431,60 +471,9 @@ class sampler:
         logging.info('Number of Walkers: ' + str(self.nwalkers))
         logging.info('Number of Tuning Generations: ' + str(len(self.mus)))
         logging.info('Scale Factor: ' + str(round(self.mu,6)))
-        logging.info('Mean Integrated Autocorrelation Time: ' + str(round(np.mean(self.autocorr_time),2)))
+        logging.info('Mean Integrated Autocorrelation Time: ' + str(round(np.mean(self.act),2)))
         logging.info('Effective Sample Size: ' + str(round(self.ess,2)))
         logging.info('Number of Log Probability Evaluations: ' + str(self.ncall))
         logging.info('Effective Samples per Log Probability Evaluation: ' + str(round(self.efficiency,6)))
         if self.thin > 1:
             logging.info('Thinning rate: ' + str(self.thin))
-
-
-    def get_chain(self, flat=False, thin=1, discard=0):
-        """
-        Get the Markov chain containing the samples.
-
-        Args:
-            flat (bool) : If True then flatten the chain into a 2D array by combining all walkers (default is False).
-            thin (int) : Thinning parameter (the default value is 1).
-            discard (int) : Number of burn-in steps to be removed from each walker (default is 0).
-
-        Returns:
-            Array object containg the Markov chain samples (2D if flat=True, 3D if flat=False).
-        """
-        if flat:
-            return self.flatten(discard=discard, thin=thin)
-        else:
-            return self.chain[discard::thin,:,:]
-
-
-    def get_log_prob(self, flat=False, thin=1, discard=0):
-        """
-        Get the value of the log probability function evalutated at the samples of the Markov chain.
-
-        Args:
-            flat (bool) : If True then flatten the chain into a 1D array by combining all walkers (default is False).
-            thin (int) : Thinning parameter (the default value is 1).
-            discard (int) : Number of burn-in steps to be removed from each walker (default is 0).
-
-        Returns:
-            Array containing the value of the log probability at the samples of the Markov chain (1D if flat=True, 2D otherwise).
-        """
-        if flat:
-            return self.samples.flatten_logprob(discard=discard, thin=thin)
-        else:
-            return self.samples.logprob[discard::thin,:]
-
-
-    @property
-    def get_last_sample(self):
-        """
-            Return the last position of the walkers.
-        """
-        return self.chain[-1]
-
-
-    def run_mcmc(self, *args, **kwargs):
-        """
-        Wrapper for run method.
-        """
-        return self.run(*args, **kwargs)
